#******************************************************************************
#
#       Copyright:      2006 Paul Obermeier (obermeier@poSoft.de)
#
#                       See the file "Tcl3D_License.txt" for information on
#                       usage and redistribution of this file, and for a
#                       DISCLAIMER OF ALL WARRANTIES.
#
#       Module:         Tcl3D -> tcl3dUtil
#       Filename:       tcl3dConsole.tcl
#
#       Author:         Paul Obermeier
#
#       Description:    Tcl module implementing a simple Tk console. It is
#                       used in the tcl3dsh, the Tcl3D Starpack.
#                       The implementation of this console window was taken
#                       from D. Richard Hipp's mktclapp.
#
#******************************************************************************

proc _tcl3dAddMenuCmd { menu label acc cmd } {
    $menu add command -label $label -accelerator $acc -command $cmd
}

proc tcl3dConsoleCreate {w prompt title} {
    upvar #0 $w.t v
    if {[winfo exists $w]} {destroy $w}
    if {[info exists v]} {unset v}
    toplevel $w
    wm title $w $title
    wm iconname $w $title
    set mnu $w.mb
    menu $mnu -borderwidth 2 -relief sunken
    $mnu add cascade -menu $mnu.file -label File -underline 0
    $mnu add cascade -menu $mnu.edit -label Edit -underline 0

    set fileMenu $mnu.file
    set editMenu $mnu.edit

    menu $fileMenu -tearoff 0
    _tcl3dAddMenuCmd $fileMenu "Save as ..." "" "tcl3dConsoleSaveFile $w.t"
    _tcl3dAddMenuCmd $fileMenu "Close"       "" "destroy $w"
    _tcl3dAddMenuCmd $fileMenu "Quit"        "" "exit"

    tcl3dConsoleCreateChild $w $prompt $editMenu
    $w configure -menu $mnu
}

proc tcl3dConsoleCreateChild {w prompt editMenu} {
    upvar #0 $w.t v
    if {$editMenu!=""} {
        menu $editMenu -tearoff 0
        _tcl3dAddMenuCmd $editMenu "Cut"   "" "tcl3dConsoleCut $w.t"
        _tcl3dAddMenuCmd $editMenu "Copy"  "" "tcl3dConsoleCopy $w.t"
        _tcl3dAddMenuCmd $editMenu "Paste" "" "tcl3dConsolePaste $w.t"
        _tcl3dAddMenuCmd $editMenu "Clear" "" "tcl3dConsoleClear $w.t"
        $editMenu add separator
        _tcl3dAddMenuCmd $editMenu "Source ..." "" "tcl3dConsoleSourceFile $w.t"
        catch {$editMenu config -postcommand "tcl3dConsoleEnableEditMenu $w"}
    }
    scrollbar $w.sb -orient vertical -command "$w.t yview"
    pack $w.sb -side right -fill y
    text $w.t -font fixed -yscrollcommand "$w.sb set"
    pack $w.t -side right -fill both -expand 1
    bindtags $w.t Console
    set v(editmenu) $editMenu
    set v(text) $w.t
    set v(history) 0
    set v(historycnt) 0
    set v(current) -1
    set v(prompt) $prompt
    set v(prior) {}
    set v(plength) [string length $v(prompt)]
    set v(x) 0
    set v(y) 0
    $w.t mark set insert end
    $w.t tag config ok -foreground blue
    $w.t tag config err -foreground red
    $w.t insert end $v(prompt)
    $w.t mark set out 1.0
    catch {rename puts tcl3dConsoleOldPuts$w}
    proc puts args [format {
        if {![winfo exists %s]} {
            rename puts {}
            rename tcl3dConsoleOldPuts%s puts
            return [uplevel #0 puts $args]
        }
        switch -glob -- "[llength $args] $args" {
            {1 *} {
                set msg [lindex $args 0]\n
                set tag ok
            }
            {2 stdout *} {
                set msg [lindex $args 1]\n
                set tag ok
            }
            {2 stderr *} {
                set msg [lindex $args 1]\n
                set tag err
            }
            {2 -nonewline *} {
                set msg [lindex $args 1]
                set tag ok
            }
            {3 -nonewline stdout *} {
                set msg [lindex $args 2]
                set tag ok
            }
            {3 -nonewline stderr *} {
                set msg [lindex $args 2]
                set tag err
            }
            default {
                uplevel #0 tcl3dConsoleOldPuts%s $args
                return
            }
        }
        tcl3dConsolePuts %s $msg $tag
    } $w $w $w $w.t]
    after idle "focus $w.t"
}

bind Console <1> {tcl3dConsoleButton1 %W %x %y}
bind Console <B1-Motion> {tcl3dConsoleB1Motion %W %x %y}
bind Console <B1-Leave> {tcl3dConsoleB1Leave %W %x %y}
bind Console <B1-Enter> {tcl3dConsoleCancelMotor %W}
bind Console <ButtonRelease-1> {tcl3dConsoleCancelMotor %W}
bind Console <KeyPress> {tcl3dConsoleInsert %W %A}
bind Console <Left> {tcl3dConsoleLeft %W}
bind Console <Control-b> {tcl3dConsoleLeft %W}
bind Console <Right> {tcl3dConsoleRight %W}
bind Console <Control-f> {tcl3dConsoleRight %W}
bind Console <BackSpace> {tcl3dConsoleBackspace %W}
bind Console <Control-h> {tcl3dConsoleBackspace %W}
bind Console <Delete> {tcl3dConsoleDelete %W}
bind Console <Control-d> {tcl3dConsoleDelete %W}
bind Console <Home> {tcl3dConsoleHome %W}
bind Console <Control-a> {tcl3dConsoleHome %W}
bind Console <End> {tcl3dConsoleEnd %W}
bind Console <Control-e> {tcl3dConsoleEnd %W}
bind Console <Return> {tcl3dConsoleEnter %W}
bind Console <KP_Enter> {tcl3dConsoleEnter %W}
bind Console <Up> {tcl3dConsolePrior %W}
bind Console <Control-p> {tcl3dConsolePrior %W}
bind Console <Down> {tcl3dConsoleNext %W}
bind Console <Control-n> {tcl3dConsoleNext %W}
bind Console <Control-k> {tcl3dConsoleEraseEOL %W}
bind Console <<Cut>> {tcl3dConsoleCut %W}
bind Console <<Copy>> {tcl3dConsoleCopy %W}
bind Console <<Paste>> {tcl3dConsolePaste %W}
bind Console <<Clear>> {tcl3dConsoleClear %W}

proc tcl3dConsolePuts {w t tag} {
    set nc [string length $t]
    set endc [string index $t [expr $nc-1]]
    if {$endc=="\n"} {
        if {[$w index out]<[$w index {insert linestart}]} {
            $w insert out [string range $t 0 [expr $nc-2]] $tag
            $w mark set out {out linestart +1 lines}
        } else {
            $w insert out $t $tag
        }
    } else {
        if {[$w index out]<[$w index {insert linestart}]} {
            $w insert out $t $tag
        } else {
            $w insert out $t\n $tag
            $w mark set out {out -1 char}
        }
    }
    $w yview insert
}

proc tcl3dConsoleInsert {w a} {
    $w insert insert $a
    $w yview insert
}

proc tcl3dConsoleLeft {w} {
    upvar #0 $w v
    scan [$w index insert] %d.%d row col
    if {$col>$v(plength)} {
        $w mark set insert "insert -1c"
    }
}

proc tcl3dConsoleBackspace {w} {
    upvar #0 $w v
    scan [$w index insert] %d.%d row col
    if {$col>$v(plength)} {
        $w delete {insert -1c}
    }
}

proc tcl3dConsoleEraseEOL {w} {
    upvar #0 $w v
    scan [$w index insert] %d.%d row col
    if {$col>=$v(plength)} {
        $w delete insert {insert lineend}
    }
}

proc tcl3dConsoleRight {w} {
    $w mark set insert "insert +1c"
}

proc tcl3dConsoleDelete w {
    $w delete insert
}

proc tcl3dConsoleHome w {
    upvar #0 $w v
    scan [$w index insert] %d.%d row col
    $w mark set insert $row.$v(plength)
}

proc tcl3dConsoleEnd w {
    $w mark set insert {insert lineend}
}

proc tcl3dConsoleEnter w {
    upvar #0 $w v
    scan [$w index insert] %d.%d row col
    set start $row.$v(plength)
    set line [$w get $start "$start lineend"]
    if {$v(historycnt)>0} {
        set last [lindex $v(history) [expr $v(historycnt)-1]]
        if {[string compare $last $line]} {
            lappend v(history) $line
            incr v(historycnt)
        }
    } else {
        set v(history) [list $line]
        set v(historycnt) 1
    }
    set v(current) $v(historycnt)
    $w insert end \n
    $w mark set out end
    if {$v(prior)==""} {
        set cmd $line
    } else {
        set cmd $v(prior)\n$line
    }
    if {[info complete $cmd]} {
        set rc [catch {uplevel #0 $cmd} res]
        if {![winfo exists $w]} return
        if {$rc} {
            $w insert end $res\n err
        } elseif {[string length $res]>0} {
            $w insert end $res\n ok
        }
        set v(prior) {}
        $w insert end $v(prompt)
    } else {
        set v(prior) $cmd
        regsub -all {[^ ]} $v(prompt) . x
        $w insert end $x
    }
    $w mark set insert end
    $w mark set out {insert linestart}
    $w yview insert
}

proc tcl3dConsolePrior w {
    upvar #0 $w v
    if {$v(current)<=0} return
    incr v(current) -1
    set line [lindex $v(history) $v(current)]
    tcl3dConsoleSetLine $w $line
}

proc tcl3dConsoleNext w {
    upvar #0 $w v
    if {$v(current)>=$v(historycnt)} return
    incr v(current) 1
    set line [lindex $v(history) $v(current)]
    tcl3dConsoleSetLine $w $line
}

proc tcl3dConsoleSetLine {w line} {
    upvar #0 $w v
    scan [$w index insert] %d.%d row col
    set start $row.$v(plength)
    $w delete $start end
    $w insert end $line
    $w mark set insert end
    $w yview insert
}

proc tcl3dConsoleButton1 {w x y} {
    global tkPriv
    upvar #0 $w v
    set v(mouseMoved) 0
    set v(pressX) $x
    set p [tcl3dConsolenearestBoundry $w $x $y]
    scan [$w index insert] %d.%d ix iy
    scan $p %d.%d px py
    if {$px==$ix} {
        $w mark set insert $p
    }
    $w mark set anchor $p
    focus $w
}

proc tcl3dConsolenearestBoundry {w x y} {
    set p [$w index @$x,$y]
    set bb [$w bbox $p]
    if {![string compare $bb ""]} {return $p}
    if {($x-[lindex $bb 0])<([lindex $bb 2]/2)} {return $p}
    $w index "$p + 1 char"
}

proc tcl3dConsoleSelectTo {w x y} {
    upvar #0 $w v
    set cur [tcl3dConsolenearestBoundry $w $x $y]
    if {[catch {$w index anchor}]} {
        $w mark set anchor $cur
    }
    set anchor [$w index anchor]
    if {[$w compare $cur != $anchor] || (abs($v(pressX) - $x) >= 3)} {
        if {$v(mouseMoved)==0} {
            $w tag remove sel 0.0 end
        }
        set v(mouseMoved) 1
    }
    if {[$w compare $cur < anchor]} {
        set first $cur
        set last anchor
    } else {
        set first anchor
        set last $cur
    }
    if {$v(mouseMoved)} {
        $w tag remove sel 0.0 $first
        $w tag add sel $first $last
        $w tag remove sel $last end
        update idletasks
    }
}

proc tcl3dConsoleB1Motion {w x y} {
    upvar #0 $w v
    set v(y) $y
    set v(x) $x
    tcl3dConsoleSelectTo $w $x $y
}

proc tcl3dConsoleB1Leave {w x y} {
    upvar #0 $w v
    set v(y) $y
    set v(x) $x
    tcl3dConsoleMotor $w
}

proc tcl3dConsoleMotor w {
    upvar #0 $w v
    if {![winfo exists $w]} return
    if {$v(y)>=[winfo height $w]} {
        $w yview scroll 1 units
    } elseif {$v(y)<0} {
        $w yview scroll -1 units
    } else {
        return
    }
    tcl3dConsoleSelectTo $w $v(x) $v(y)
    set v(timer) [after 50 tcl3dConsoleMotor $w]
}

proc tcl3dConsoleCancelMotor w {
    upvar #0 $w v
    catch {after cancel $v(timer)}
    catch {unset v(timer)}
}

proc tcl3dConsoleCopy w {
    if {![catch {set text [$w get sel.first sel.last]}]} {
        clipboard clear -displayof $w
        clipboard append -displayof $w $text
    }
}

proc tcl3dConsoleCanCut w {
    set r [catch {
        scan [$w index sel.first] %d.%d s1x s1y
        scan [$w index sel.last] %d.%d s2x s2y
        scan [$w index insert] %d.%d ix iy
    }]
    if {$r==1} {return 0}
    if {$s1x==$ix && $s2x==$ix} {return 1}
    return 2
}

proc tcl3dConsoleCut w {
    if {[tcl3dConsoleCanCut $w]==1} {
        tcl3dConsoleCopy $w
        $w delete sel.first sel.last
    }
}

proc tcl3dConsolePaste w {
    if {[tcl3dConsoleCanCut $w]==1} {
        $w delete sel.first sel.last
    }
    if {[catch {selection get -displayof $w -selection CLIPBOARD} topaste]} {
        return
    }
    set prior 0
    foreach line [split $topaste \n] {
        if {$prior} {
            tcl3dConsoleEnter $w
            update
        }
        set prior 1
        $w insert insert $line
    }
}

proc tcl3dConsoleEnableEditMenu w {
    upvar #0 $w.t v
    set m $v(editmenu)
    if {$m=="" || ![winfo exists $m]} return
    switch [tcl3dConsoleCanCut $w.t] {
        0 {
            $m entryconf Copy -state disabled
            $m entryconf Cut -state disabled
        }
        1 {
            $m entryconf Copy -state normal
            $m entryconf Cut -state normal
        }
        2 {
            $m entryconf Copy -state normal
            $m entryconf Cut -state disabled
        }
    }
}

proc tcl3dConsoleSourceFile w {
    set types {
        {{TCL Scripts}  {.tcl}}
        {{All Files}    *}
    }
    set f [tk_getOpenFile -filetypes $types -title "TCL Script To Source..."]
    if {$f!=""} {
        uplevel #0 source $f
    }
}

proc tcl3dConsoleSaveFile w {
    set types {
        {{Text Files}  {.txt}}
        {{All Files}    *}
    }
    set f [tk_getSaveFile -filetypes $types -title "Write Screen To..."]
    if {$f!=""} {
        if {[catch {open $f w} fd]} {
            tk_messageBox -type ok -icon error -message $fd
        } else {
            puts $fd [string trimright [$w get 1.0 end] \n]
            close $fd
        }
    }
}

proc tcl3dConsoleClear w {
    $w delete 1.0 {insert linestart}
}
