#******************************************************************************
#
#       Copyright:      2005-2006 Paul Obermeier (obermeier@poSoft.de)
#
#                       See the file "Tcl3D_License.txt" for information on
#                       usage and redistribution of this file, and for a
#                       DISCLAIMER OF ALL WARRANTIES.
#
#       Module:         Tcl3D -> tcl3dUtil
#       Filename:       tcl3dVector.tcl
#
#       Author:         Paul Obermeier
#
#       Description:    Tcl module to handle tcl3dVectors, i.e. contiguous
#                       pieces of memory.
#
#******************************************************************************

# Create a new Tcl3D Vector of size "size" by calling the memory allocation
# routine new_"type" and create a new Tcl procedure.

proc tcl3dVector { type size } {
    set ptr [new_$type $size]
    set code {
        set method [lindex $args 0]
        set parms [concat $ptr [lrange $args 1 end]]
        switch $method {
            get    {return [eval "${type}_getitem $parms"]}
            set    {return [eval "${type}_setitem $parms"]}
            setvec {return [eval "${type}_setarray $parms"]}
            addvec {return [eval "${type}_addarray $parms"]}
            mulvec {return [eval "${type}_mularray $parms"]}
            delete {eval "delete_$type $ptr; rename $ptr {}"}
        }
    }
    # Create a procedure
    uplevel "proc $ptr args {set ptr $ptr; set type $type;$code}"
    return $ptr
}

proc tcl3dVectorInd { vec type ind } {
    eval "${type}_ind $vec $ind" 
}

# Print the first "num" elements of Tcl3D Vector "vec" onto standard output.

proc tcl3dVectorPrint { vec num } {
    for { set i 0 } { $i < $num } { incr i } {
        puts "$i : [$vec get $i]"
    }
}

# Create a new Tcl3D Vector of type "type" from given variable argument list.

proc tcl3dVectorFromArgs { type args } {
    set len [llength $args]
    if { $len == 0 } {
        error "Vector init has zero length"
    }
    set a [tcl3dVector $type $len]
    set i 0
    foreach elem $args {
        $a set $i $elem
        incr i
    }
    return $a
}

# Create a new Tcl3D Vector of type "type" from given Tcl list "l".

proc tcl3dVectorFromList { type l } {
    set len [llength $l]
    if { $len == 0 } {
        error "List has zero length"
    }
    set a [tcl3dVector $type $len]
    set i 0
    foreach elem $l {
        $a set $i $elem
        incr i
    }
    return $a
}

# Internal helper function.

proc tcl3dCharToNum { char } {
    scan $char %c value
    return $value
}

# Create a new Tcl3D Vector of type "type" from given string "str".
# This version is very slow and was intended only for short text strings.
# Deprecated, use tcl3dVectorFromByteArray instead.

proc tcl3dVectorFromString { type str } {
    set len [string length $str]
    if { $len == 0 } {
        error "String has zero length"
    }
    set a [tcl3dVector $type $len]
    set i 0
    foreach elem [split $str ""] {
        # puts "Setting a($i) to: $elem ( [tcl3dCharToNum $elem] )"
        $a set $i [tcl3dCharToNum $elem]
        incr i
    }
    return $a
}

# Create a new Tcl3D Vector of type "type" from given binary string "str".

proc tcl3dVectorFromByteArray { type str } {
    set len [string length $str]
    if { $len == 0 } {
        error "String has zero length"
    }
    set a [tcl3dVector $type $len]
    tcl3dByteArray2Vector $str $a $len 0 0
    return $a
}

# Create a new Tcl3D Vector containing the image data of Tk photo phImg.
# Only GL_UNSIGNED_BYTE currently supported.

proc tcl3dVectorFromPhoto { phImg { numChans -1 } { scl 1 } { off 0 }  } {
    set w [image width  $phImg]
    set h [image height $phImg]
    set n [tcl3dPhotoChans $phImg]
    if { $numChans > $n || $numChans <= 0 } {
        set numChans $n
    }
    set texImg [tcl3dVector GLubyte [expr {$w * $h * $numChans}]]
    tcl3dPhotoToVector $phImg $texImg $numChans $scl $off
    return $texImg
}

# Copy "num" elements of Tcl3D Vector "vec" into a Tcl list and return that list.

proc tcl3dVectorToList { vec num } {
    set l {}
    for { set i 0 } { $i < $num } { incr i } {
        set val [$vec get $i]
        lappend l $val
    }
    return $l
}

# Interpret the elements of Tcl3D Vector "vec" (which must be of type GLubyte) as a
# null-terminated string and return that string.
# This version is very slow and was intended only for short text strings.
# Deprecated, use tcl3dVectorToByteArray instead.

proc tcl3dVectorToString { vec } {
    set i 0
    set str ""
    while { 1 } {
        set val [$vec get $i]
        if { $val == 0 } {
            break
        }
        append str $val
        incr i
    }
    return $str
}

# Copy "numBytes" elements of Tcl3D Vector "vec" into a Tcl binary string and return
# that string. The Tcl3D Vector has be of type GLubyte.

proc tcl3dVectorToByteArray { vec numBytes {srcOff 0} {destOff 0} } {
    if { $numBytes == 0 } {
        error "Vector has zero length"
    }
    # First generate a binary string of approriate size.
    set tmp [binary format c 0]
    set bytes [string repeat $tmp $numBytes]
    tcl3dVector2ByteArray $vec $bytes $numBytes $srcOff $destOff
    return $bytes
}
